# pylint: disable-msg=line-too-long
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Union
from warnings import warn
import copy
import os


__all__ = [
    "Step",
    "Configuration",
    "set_project_root",
    "get_project_root"
]


class Step:
    """
    An instance of this class defines a single launch of an external command by Universum. It determines execution
    parameters and result handling. Individual build steps are collected in a :class:`Configuration` object, that
    allows to create linear and nested step structures.
    See https://universum.readthedocs.io/en/latest/configuring.html#build-step for details.

    **Step keys**

    Parameters supplied via named attributes during construction are annotated and can be type-checked for safety.
    Users may also define custom parameters in `kwargs` - these parameters will be stored internally in `dict` fields
    and can be retrieved by indexing. The following list enumerates all the keys used by Universum:

    name
        The human-readable name of a build step. The name is used as the title of the build log block
        corresponding to the execution of this step. It is also used to generate name of the log file if the option
        for storing to log files is selected. If several build steps have the same names, and logs are stored to
        files (see ``--out`` / ``-o``  command-line parameter for details), all logs for such
        steps will be stored to one file in order of their appearances.
    command
        The command line for the build step launch. For every step the first list
        item should be a console command (e.g. script name, or any other command like ``ls``), and other list items
        should be the arguments, added to the command (e.g. ``--debug`` or ``-la``). Every command line element,
        separated with space character, should be passed as a separate string argument. Lists like ``["ls -a"]``
        will not be processed correctly and therefore should be splat into ``["ls", "-a"]``. Lists like
        ``["build.sh", "--platform A"]`` will not be processed correctly and thus should be plat into
        ``["build.sh", "--platform", "A"]``. A build step can have an empty list as a command - such step won't do
        anything except showing up the step name in Universum execution logs. Some common actions, such as ``echo``,
        are bash features and not actual programs to run. These features should be called as
        ``["bash", "-c", "echo -e 'Some line goes here'"]``. Any other shell can also be used instead of bash.
        Note that in this case the string to be passed to bash is one argument containing white spaces
        and therefore not splat by commas.

        .. note::

            When a process is launched via shell, that shell also parses the arguments, which does not happen
            with arguments of a direct process launch. In case of direct process launch the arguments including
            special characters (e.g. ``*.txt``) are treated as usual strings. To make the lack of filename expansion
            (globbing) more obvious and to make the relaunch of executed command by copy-pasting more convenient,
            these arguments are printed within quotes in log. To use bash for globbing, please also use ``bash -c``
            as explained above.

    environment
        Required environment variables, e.g. ``environment={"VAR1": "String", "VAR2": "123"}`` Can be set at any
        step level, but re-declaring variables is not supported, so please make sure to mention every variable only
        one time at most.
    artifacts
        Path to the file or directory to be copied to the working directory as an execution
        result. Can contain shell-style pattern matching (e.g. `"out/*.html"`), including recursive wildcards
        (e.g. `"out/**/index.html"`). If not stated otherwise (see ``--no-archive``
        command-line parameter for details), artifact directories are copied
        as archives. If `artifact_prebuild_clean` key is either absent or set to `False`
        and stated artifacts are present in downloaded sources, it is considered a failure and configuration
        execution will not proceed. If no required artifacts were found in the end of the `Universum` run, it is
        also considered a failure. In case of shell-style patterns build is failed if no files or directories
        matching pattern are found.
    report_artifacts
        Path to the special artifacts for reporting (e.g. to Swarm). Unlike `artifacts` key, `report_artifacts`
        are not obligatory and their absence is not considered a build failure. A directory cannot be stored
        as a separate artifact, so when using ``--no-archive`` option, do not claim directories as `report_artifacts`.
        Please note that any file can be put as `artifact`, `report_artifact`, or both. A file that is both
        in `artifacts` and `report_artifacts`, will be mentioned in a report and will cause build failure when missing.
    artifact_prebuild_clean
        A flag to signal that artifacts must be cleaned before the build. Cleaning of single files, directories and
        sets of files defined by shell-style patterns is supported. By default, artifacts are not stored in VCS, and
        artifact presence before build most likely means that working directory is not cleaned after previous build and
        therefore might influence build results. But sometimes deliverables have to be stored in VCS, and in this case
        instead of stopping the build they should be simply cleaned before it. This is where
        ``artifact_prebuild_clean=True`` key is supposed to be used. This flag is ignored, if both `artifacts` and
        `report_artifacts` are not set.
    directory
        Path to a current working directory for launched process. Absent
        `directory` has equal meaning to empty string passed as a `directory` value and means that `command`
        will be launched from the project root directory. See
        https://universum.readthedocs.io/en/latest/configuring.html#execution-directory for details.
    critical
        A flag used in case of a linear step execution, when the result of some step is critical
        for the subsequent step execution. If some step has `critical` key set to `True` and executing this step
        fails, no more steps will be executed during this run. However, all background steps, which have already
        started will be finished regardless of critical step results.
    background
        A flag used to signal that the current step should be executed independently in parallel with
        all other steps. All logs from such steps are written to file, and the results of execution are collected
        in the end of `Universum` run. Next step execution begins immediately after starting a background step,
        not waiting for it to be completed. Several background steps can be executed simultaneously.
    finish_background
        A flag used to signal that the current step should be executed only after ongoing
        background steps (if any) are finished.
    code_report
        A flag used to signal that the current step performs static or syntax analysis of the code.
        Usually set in conjunction with adding ``--result-file="${CODE_REPORT_FILE}"`` to 'command' arguments.
        Analyzers currently provided by Universum are: ``pylint``, ``svace`` and ``uncrustify``.
        See https://universum.readthedocs.io/en/latest/code_report.html for details.
    pass_tag
        A tag used to mark successful TeamCity builds. This tag can be set independenty
        of `fail_tag` value per each step. The value should be set to a strings without spaces as acceptable by
        TeamCity as tags. Every tag is added (if matching condition) after executing build step it is set in,
        not in the end of all run.
    fail_tag
        A tag used to mark failed TemCity builds. See `pass_tag` for details.

    Each parameter is optional, and is substituted with a falsy value, if omitted.

    :Example of a simple Step construction:

    >>> Step(foo='bar')
    {'foo': 'bar'}
    >>> step = Step(name='foo', command=['1', '2', '3'], _extras={'_extras': 'test', 'bool': True}, myvar=1)
    >>> step
    {'name': 'foo', 'command': ['1', '2', '3'], '_extras': {'_extras': 'test', 'bool': True}, 'myvar': 1}
    >>> step['name']
    'foo'
    >>> step['_extras']
    {'_extras': 'test', 'bool': True}
    >>> step['myvar']
    1

    :Example of using Configuration objects to wrap individual build steps:

    >>> make = Configuration([Step(name="Make ", command=["make"], pass_tag="pass_")])
    >>> target = Configuration([
    ...              Step(name="Linux", command=["--platform", "Linux"], pass_tag="Linux"),
    ...              Step(name="Windows", command=["--platform", "Windows"], pass_tag="Windows")
    ...          ])
    >>> configs = make * target
    >>> configs.dump()
    "[{'name': 'Make Linux', 'command': 'make --platform Linux', 'pass_tag': 'pass_Linux'},\\n{'name': 'Make Windows', 'command': 'make --platform Windows', 'pass_tag': 'pass_Windows'}]"

    This means that tags "pass_Linux" and "pass_Windows" will be sent to TeamCity's build.

    .. note::

        All the paths, specified in `command`, `artifacts` and `directory` parameters, can be absolute or
        relative. All relative paths start from the project root (see :ref:`get_project_root`).
    """

    # pylint: disable-msg=too-many-locals
    def __init__(self,
                 name: str = '',
                 command: Optional[List[str]] = None,
                 environment: Optional[Dict[str, str]] = None,
                 artifacts: str = '',
                 report_artifacts: str = '',
                 artifact_prebuild_clean: bool = False,
                 directory: str = '',
                 critical: bool = False,
                 background: bool = False,
                 finish_background: bool = False,
                 code_report: bool = False,
                 pass_tag: str = '',
                 fail_tag: str = '',
                 if_env_set: str = '',
                 **kwargs) -> None:
        self.name: str = name
        self.directory: str = directory
        self.code_report: bool = code_report
        self.command: List[str] = command if command else []
        self.environment: Dict[str, str] = environment if environment else {}
        self.artifacts: str = artifacts
        self.report_artifacts: str = report_artifacts
        self.artifact_prebuild_clean: bool = artifact_prebuild_clean
        self.critical: bool = critical
        self.background: bool = background
        self.finish_background: bool = finish_background
        self.pass_tag: str = pass_tag
        self.fail_tag: str = fail_tag
        self.if_env_set: str = if_env_set
        self.children: Optional['Configuration'] = None
        self._extras: Dict[str, str] = {}
        for key, value in kwargs.items():
            self._extras[key] = value

    def __repr__(self) -> str:
        """
        This function simulates `dict`-like representation for string output. This is useful when printing contents of
        :class:`Configuration` objects, as internally they wrap a list of :class:`Step` objects

        :return: a `dict`-like string

        >>> step = Step(name='foo', command=['bar'], my_var='baz')
        >>> repr(step)
        "{'name': 'foo', 'command': 'bar', 'my_var': 'baz'}"
        """
        res = {k: v for k, v in self.__dict__.items() if v and k != '_extras'}
        res.update(self._extras)
        if len(self.command) == 1:  # command should be printed as one string, instead of list
            res['command'] = self.command[0]
        return str(res)

    def __eq__(self, other: Any) -> bool:
        """
        This functions simulates `dict`-like check for match

        :param other: `dict` to compare values, or :class:`Step` object to check equality
        :return: `True` if `other` matches

        >>> step1 = Step(name='foo', my_var='bar')
        >>> step2 = Step(name='foo', my_var='bar')
        >>> step3 = Step(name='foo', my_var='bar', critical=True)
        >>> step1 == step1
        True
        >>> step1 == step2
        True
        >>> step1 == step3
        False
        >>> step1 == {'name': 'foo', 'my_var': 'bar'}
        True
        >>> step1 == {'name': 'foo', 'my_var': 'bar', 'critical': False}
        True
        >>> step1 == {'name': 'foo', 'my_var': 'bar', 'test': None}
        True
        >>> step1 == {'name': 'foo', 'my_var': 'bar', 'test': ''}
        True
        >>> step1 == {'name': 'foo', 'my_var': 'bar', 'test': ' '}
        False
        """
        if isinstance(other, Step):
            return self == other.__dict__
        if isinstance(other, dict):
            for key, val in other.items():
                if val and self[key] != val:
                    return False
            return True
        return super().__eq__(other)

    def __getitem__(self, key: str) -> Any:
        """
        This functions simulates `dict`-like legacy read access

        .. note::

            It is recommended to use field-like access to non user-defined attributes of Step.
            This preserves data type information for mypy static analysis.

        :param key: client-defined item
        :return: client-defined value

        >>> step = Step(name='foo', my_var='bar')
        >>> step['name'] == step.name == 'foo'
        True

        Note that step['name'] has type 'Any', but step.name has type 'str'

        >>> step['my_var']
        'bar'
        >>> step['test']
        """
        #  _extras are checked first - just in case _extras field is added manually
        # do note that __setitem__ checks predefined fields first, however it's impossible to shadow them by
        # modifying _extras
        return self._extras.get(key, self.__dict__.get(key, None))

    def __setitem__(self, key: str, value: Any) -> None:
        """
        This functions simulates `dict`-like legacy write access

        .. note::

            It is recommended to use field-like access to non user-defined attributes of Step.
            This allows static analysis to catch possible type mismatch.

        :param key: client-defined key
        :param value: client-defined value

        >>> import warnings
        >>> def do_and_get_warnings(f):
        ...     with warnings.catch_warnings(record=True) as w:
        ...         warnings.simplefilter("always")
        ...         f()
        ...         return w
        >>> def assign_legacy(o, k, v):
        ...     o[k] = v
        >>> def assign_name_legacy(o, v):
        ...     o['name'] = v
        >>> def assign_name_new(o, v):
        ...     o.name = v
        >>> step = Step(name='foo', my_var='bar')
        >>> do_and_get_warnings(lambda : assign_legacy(step, 'name', 'bar'))  # doctest: +ELLIPSIS
        [<warnings.WarningMessage object at ...>]
        >>> step['name']
        'bar'
        >>> do_and_get_warnings(lambda : assign_name_new(step, 'baz'))  # doctest: +ELLIPSIS
        []
        >>> step['name']
        'baz'
        >>> do_and_get_warnings(lambda : assign_legacy(step, 'directory', 'foo'))  # doctest: +ELLIPSIS
        [<warnings.WarningMessage object at ...>]
        >>> do_and_get_warnings(lambda : assign_legacy(step, 'test', 42))
        []
        >>> do_and_get_warnings(lambda : assign_legacy(step, '_extras', {'name': 'baz'}))
        []
        >>> step
        {'name': 'baz', 'directory': 'foo', 'my_var': 'bar', 'test': 42, '_extras': {'name': 'baz'}}
        """
        if key in self.__dict__ and key != '_extras':
            warn("Re-defining the value of Step field. Please use var." + key + " to set it instead of "
                 "using var['" + key + "']")
            self.__dict__[key] = value
        else:
            self._extras[key] = value

    def get(self, key: str, default: Any = None) -> Any:
        """
        This functions simulates `dict`-like legacy read access

        .. note::

            It is recommended to use field-like access to non user-defined attributes of Step.
            This preserves data type information for mypy static analysis.

        :param key: client-defined item
        :param default: value to return if `key` is absent in `dict`
        :return: client-defined value

        >>> import warnings
        >>> def do_and_get_warnings(f):
        ...     with warnings.catch_warnings(record=True) as w:
        ...         warnings.simplefilter("always")
        ...         f()
        ...         return w
        >>> step = Step(name='foo', my_var='bar')
        >>> do_and_get_warnings(lambda : step.get('name', 'test'))  # doctest: +ELLIPSIS
        [<warnings.WarningMessage object at ...>]

        Note that step.get('name') has type 'Any', but step.name has type 'str'

        >>> step.get('my_var', 'test')
        'bar'
        >>> step.get('my_var_2', 'test')
        'test'
        >>> step.get('command', 'test')
        'test'
        """
        result = self._extras.get(key)
        if result:
            return result
        result = self.__dict__.get(key)
        if result:
            warn("Using legacy API to access configuration values. Please use var." + key + " instead.")
            return result
        return default

    def __add__(self, other: 'Step') -> 'Step':
        """
        This functions defines operator ``+`` for :class:`Step` class objects by
        concatenating strings and contents of dictionaries. Note that `critical` attribute is not merged.

        :param other: `Step` object
        :return: new `Step` object, including all attributes from both `self` and `other` objects

        >>> step1 = Step(name='foo', command=['foo'], critical=True, my_var1='foo')
        >>> step2 = Step(name='bar', command=['bar'], background=True, my_var1='bar', my_var2='baz')
        >>> step1 + step2
        {'name': 'foobar', 'command': ['foo', 'bar'], 'background': True, 'my_var1': 'foobar', 'my_var2': 'baz'}
        """
        return Step(
            name=self.name + other.name,
            command=self.command + other.command,
            environment=combine(self.environment, other.environment),
            artifacts=self.artifacts + other.artifacts,
            report_artifacts=self.report_artifacts + other.report_artifacts,
            artifact_prebuild_clean=self.artifact_prebuild_clean or other.artifact_prebuild_clean,
            directory=self.directory + other.directory,
            critical=False,
            background=self.background or other.background,
            finish_background=self.finish_background or other.finish_background,
            code_report=self.code_report or other.code_report,
            pass_tag=self.pass_tag + other.pass_tag,
            fail_tag=self.fail_tag + other.fail_tag,
            if_env_set=self.if_env_set + other.if_env_set,
            **combine(self._extras, other._extras)
        )

    def replace_string(self, from_string: str, to_string: str) -> None:
        """
        Replace instances of a string, used for pseudo-variables

        :param from_string: string to replace, e.g. `${CODE_REPORT_FILE}`
        :param to_string: value to put in place of `from_string`

        >>> step = Step(name='foo test', command=['foo', 'baz', 'foobar'], myvar1='foo', myvar2='bar', myvar3=1)
        >>> step.replace_string('foo', 'bar')
        >>> step
        {'name': 'foo test', 'command': ['bar', 'baz', 'barbar'], 'myvar1': 'bar', 'myvar2': 'bar', 'myvar3': 1}

        >>> step = Step(artifacts='foo', report_artifacts='foo', directory='foo')
        >>> step.replace_string('foo', 'bar')
        >>> step
        {'directory': 'bar', 'artifacts': 'bar', 'report_artifacts': 'bar'}
        """
        self.command = [word.replace(from_string, to_string) for word in self.command]
        self.artifacts = self.artifacts.replace(from_string, to_string)
        self.report_artifacts = self.report_artifacts.replace(from_string, to_string)
        self.directory = self.directory.replace(from_string, to_string)
        for k, v in self._extras.items():
            if isinstance(v, str):
                self._extras[k] = v.replace(from_string, to_string)

    def stringify_command(self) -> bool:
        """
        Concatenates components of a command into one element

        :return: `True`, if any of the command components have space inside

        >>> step = Step(name='stringify test', command=['foo', 'bar', '--baz'])
        >>> step.stringify_command()
        False
        >>> step
        {'name': 'stringify test', 'command': 'foo bar --baz'}
        >>> step.stringify_command()
        True
        """
        result = False
        command_line = ""
        for argument in self.command:
            if " " in argument:
                argument = "\"" + argument + "\""
                result = True
            command_line = command_line + " " + argument if command_line else argument
        self.command = [command_line] if command_line else []
        return result


DictType = TypeVar('DictType', bound=dict)


def combine(dictionary_a: DictType, dictionary_b: DictType) -> DictType:
    # TODO: move to utils, as this is no longer specific to configurations
    """
    Combine two dictionaries using plus operator for matching keys

    :param dictionary_a: may have any keys and values
    :param dictionary_b: may have any keys, but the values of keys, matching `dictionary_a`,
        should be the compatible, so that `dictionary_a[key] + dictionary_b[key]` is a valid expression
    :return: new dictionary containing all keys from both `dictionary_a` and `dictionary_b`;
        for each matching key the value in resulting dictionary is a sum of two corresponding values

    For example:

    >>> combine(dict(attr_a = "a1", attr_b = ["b11", "b12"]), dict(attr_a = "a2", attr_b = ["b2"]))
    {'attr_a': 'a1a2', 'attr_b': ['b11', 'b12', 'b2']}

    >>> combine(dict(attr_a = {"a1": "v1"}, attr_b = {"b1": "v1"}), dict(attr_a = {"a2": "v2"}))
    {'attr_a': {'a1': 'v1', 'a2': 'v2'}, 'attr_b': {'b1': 'v1'}}

    """

    result = dictionary_a.__class__()
    for key in dictionary_a:
        if key in dictionary_b:
            if isinstance(dictionary_a[key], dict) and isinstance(dictionary_b[key], dict):
                # TODO: shouldn't we use recursion here?
                # also why not write result[key] = {**dictionary_a[key], **dictionary_b[key]} ?
                new_value = dictionary_a[key].copy()
                new_value.update(dictionary_b[key])
                result[key] = new_value
            else:
                result[key] = dictionary_a[key] + dictionary_b[key]
        else:
            result[key] = dictionary_a[key]

    for key in dictionary_b:
        if key not in dictionary_a:
            result[key] = dictionary_b[key]

    return result


class Configuration:
    """
    Configuration is a class for establishing project configurations. Each Configuration object wraps a list of build
    steps either from a pre-constructed :class:`Step` objects, or from the supplied `dict` data.

    >>> cfg1 = Configuration([{"field1": "string"}])
    >>> cfg1.configs
    [{'field1': 'string'}]

    Build-in method :meth:`.all()` generates iterable for all configuration dictionaries for further usage:

    >>> for i in cfg1.all(): i
    {'field1': 'string'}

    Built-in method :meth:`.dump()` will generate a printable string representation of the object.
    This string will be printed into console output

    >>> cfg1.dump()
    "[{'field1': 'string'}]"

    Adding two objects will extend list of dictionaries:

    >>> cfg2 = Configuration([{"field1": "line"}])
    >>> for i in (cfg1 + cfg2).all(): i
    {'field1': 'string'}
    {'field1': 'line'}

    While multiplying them will combine same fields of dictionaries:

    >>> for i in (cfg1 * cfg2).all(): i
    {'field1': 'stringline'}

    When a field value is a list itself -

    >>> cfg3 = Configuration([dict(field2=["string"])])
    >>> cfg4 = Configuration([dict(field2=["line"])])

    multiplying them will extend the inner list:

    >>> for i in (cfg3 * cfg4).all(): i
    {'field2': ['string', 'line']}

    """

    def __init__(self, lst: Optional[Union[List[Dict[str, Any]], List[Step]]] = None):
        #  lst can be List[Dict[str, Any]] to support legacy cases - should be removed after project migration
        self.configs: List[Step] = []  # aggregation is used instead of inheritance for type safety
        if lst:
            for item in lst:
                if isinstance(item, Step):
                    self.configs.append(item)
                else:
                    self.configs.append(Step(**item))

    def __eq__(self, other: Any) -> bool:
        """
        This function checks wrapped configurations for match

        :param other: :class:`Configuration` object or list of :class:`Step` objects
        :return: `True` if stored configurations match

        >>> l = [{'name': 'foo', 'critical': True}, {'name': 'bar', 'myvar': 'baz'}]
        >>> v1 = Configuration(l)
        >>> step1 = Step(name='foo', critical=True)
        >>> step2 = Step(name='bar', myvar='baz')
        >>> v2 = Configuration([step1]) + Configuration([step2])
        >>> v3 = Configuration() + Configuration([l[0]]) + Configuration([step2])
        >>> v1 == v1
        True
        >>> v1 == v2
        True
        >>> v1 == v3
        True
        >>> v1 == v1 + Configuration()
        True
        >>> v1 == v1 * 1
        True
        >>> v1 == v1 * Configuration()
        True
        >>> v1 == v2 + v3
        False
        """
        if isinstance(other, Configuration):
            return self.configs == other.configs
        if isinstance(other, list):
            return self.configs == other
        return super().__eq__(other)

    def __bool__(self) -> bool:
        """
        This function defines truthiness of :class:`Configuration` object

        :return: `True` if :class:`Configuration` class object is empty

        >>> cfg1 = Configuration()
        >>> bool(cfg1)
        False
        >>> cfg2 = Configuration([])
        >>> bool(cfg2)
        False
        >>> cfg3 = Configuration([{}])
        >>> bool(cfg3)
        True
        """
        return len(self.configs) != 0

    def __getitem__(self, item: int) -> Step:
        return self.configs[item]

    def __add__(self, other: Union['Configuration', List[Step]]) -> 'Configuration':
        """
        This functions defines operator ``+`` for :class:`Configuration` class objects by
        concatenating lists of dictionaries into one list.
        The order of list members in resulting list is preserved: first all dictionaries from `self`,
        then all dictionaries from `other`.

        :param other: `Configuration` object OR list of project configurations to be added to `self`
        :return: new `Configuration` object, including all configurations from both `self` and `other` objects
        """
        list_other: List[Step] = other.configs if isinstance(other, Configuration) else other
        return Configuration(list.__add__(self.configs, list_other))

    def __mul__(self, other: Union['Configuration', int]) -> 'Configuration':
        """
        This functions defines operator ``*`` for :class:`Configuration` class objects.
        The resulting object is created by combining every `self` list member with
        every `other` list member using :func:`.combine()` function.

        :param other: `Configuration` object  OR an integer value to be multiplied to `self`
        :return: new `Configuration` object, consisting of the list of combined configurations
        """
        if isinstance(other, int):
            return Configuration(list.__mul__(self.configs, other))
        config_list: List[Step] = []
        for obj_a in self.configs:
            obj_a_copy = copy.deepcopy(obj_a)
            if obj_a.children:
                obj_a_copy.children = obj_a.children * other
            else:
                obj_a_copy.children = copy.deepcopy(other)

            config_list.append(obj_a_copy)
        return Configuration(config_list)

    def all(self) -> Iterable[Step]:
        """
        Function for configuration iterating.

        :return: iterable for all dictionary objects in :class:`Configuration` list
        """
        for obj_a in self.configs:
            if obj_a.children:
                for obj_b in obj_a.children.all():
                    yield obj_a + obj_b
            else:
                yield copy.deepcopy(obj_a)

    def dump(self, produce_string_command: bool = True) -> str:
        """
        Function for :class:`Configuration` objects pretty printing.

        :param produce_string_command: if set to False, prints "command" as list instead of string
        :return: a user-friendly string representation of all configurations list
        """
        space_found: bool = False
        result: str = "["
        for obj in self.all():
            if len(result) > 1:
                result += ",\n"

            if produce_string_command and obj.stringify_command():
                space_found = True

            result += str(obj)

        result += "]"

        if space_found:
            result += "\n\nWARNING! We have detected space character within some of the command-line parameters.\n"
            result += "Please make sure you are not trying to pass two or more parameters as one."
        return result

    def filter(self, checker: Callable[[Step], bool],
               parent: Step = None) -> 'Configuration':
        """
        This function is supposed to be called from main script, not configuration file.
        It uses provided `checker` to find all the configurations that pass the check,
        removing those not matching conditions.

        :param checker: a function that returns `True` if configuration passes the filter and `False` otherwise
        :param parent: an inner parameter for recursive usage; should be None when function is called from outside
        :return: new `Configuration` object without configurations not matching `checker` conditions
        """

        if parent is None:
            parent = Step()

        filtered_configs: Configuration = Configuration()

        for obj_a in self.configs:
            item: Step = parent + obj_a

            if obj_a.children:
                active_children_configs = obj_a.children.filter(checker, item).configs
                if active_children_configs:
                    if len(active_children_configs) == 1:
                        obj_a_copy = obj_a + active_children_configs[0]
                        if active_children_configs[0].children:
                            obj_a_copy.children = active_children_configs[0].children
                        obj_a_copy.critical = \
                            obj_a.critical or active_children_configs[0].critical
                    else:
                        obj_a_copy = copy.deepcopy(obj_a)
                        obj_a_copy.children = Configuration(active_children_configs)
                    filtered_configs += [obj_a_copy]
            else:
                if checker(item):
                    filtered_configs += [copy.deepcopy(obj_a)]

        return filtered_configs


global_project_root = os.getcwd()


#: Variations is preserved as legacy alias for :class:`Configuration`
Variations = Configuration


def set_project_root(project_root: str) -> None:
    """
    Function to be called from main script; not supposed to be used in configuration file.
    Stores generated project location for further usage. This function is needed because
    project sources most likely will be copied to a temporary directory
    of some automatically generated location and the CI run will be performed there.

    :param project_root: path to actual project root
    """
    global global_project_root
    global_project_root = project_root


def get_project_root() -> str:
    """
    Function to be used in configuration file. Inserts actual project location after
    that location is generated. If project root is not set, function returns current directory.

    :return: actual project root
    """
    return os.path.abspath(global_project_root)



import os

from ..lib import utils
from ..lib.gravity import Module


class ProjectDirectory(Module):
    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Source files",
                                                     "Parameters determining the processing of repository files")

        parser.add_argument("--project-root", "-pr", dest="project_root", metavar="PROJECT_ROOT",
                            help="Temporary directory to copy sources to. Default is 'temp'")

    def __init__(self, *args, **kwargs):
        if self.settings.project_root:
            self.settings.project_root = utils.parse_path(self.settings.project_root, os.getcwd())
        else:
            self.settings.project_root = os.path.join(os.getcwd(), "temp")
        super().__init__(*args, **kwargs)



import sys
from typing import List
import inspect

from ..lib.gravity import Module, Dependency


class GlobalErrorState(Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        self.errors: List[str] = []

    def get_errors(self) -> List[str]:
        return self.errors

    def is_in_error_state(self) -> bool:
        return len(self.errors) > 0


class HasErrorState(Module):
    global_error_state_factory = Dependency(GlobalErrorState)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        self.global_error_state: GlobalErrorState = self.global_error_state_factory()

    def error(self, message: str) -> None:
        self.global_error_state.errors.append(inspect.cleandoc(message))

    def is_in_error_state(self) -> bool:
        return self.global_error_state.is_in_error_state()

    def check_required_option(self, setting_name: str, message: str) -> bool:
        if not getattr(self.settings, setting_name, None):
            self.error(message)
            return False
        return True

    def read_and_check_multiline_option(self, setting_name: str, error_message: str) -> str:
        try:
            value: str = getattr(self.settings, setting_name, None)
            if value.startswith('@'):
                try:
                    with open(value.lstrip('@')) as file_name:
                        result = file_name.read()
                except FileNotFoundError as e:
                    self.error(f"Error reading argument {setting_name} from file {e.filename}: no such file")
                    return ""

            elif value == '-':
                result = "".join(sys.stdin.readlines())
            else:
                result = value
        except AttributeError as e:
            self.error(error_message)
            return ""

        if not result:
            self.error(error_message)
            return ""

        return result



from collections import defaultdict

from ..lib.gravity import Module, Dependency
from ..lib.utils import make_block
from . import automation_server
from .output import HasOutput
from .structure_handler import HasStructure

__all__ = [
    "ReportObserver",
    "Reporter"
]


class ReportObserver:
    """
    Abstract base class for reporting modules
    """

    def get_review_link(self):
        raise NotImplementedError

    def report_start(self, report_text):
        raise NotImplementedError

    def report_result(self, result, report_text=None, no_vote=False):
        raise NotImplementedError

    def code_report_to_review(self, report):
        raise NotImplementedError


class Reporter(HasOutput, HasStructure):
    automation_server_factory = Dependency(automation_server.AutomationServerForHostingBuild)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Result reporting",
                                                     "Build results collecting and publishing parameters")

        parser.add_argument("--report-build-start", "-rst", action="store_true", dest="report_start",
                            help="Send additional comment to review system on build started (with link to log)")
        parser.add_argument("--report-build-success", "-rsu", action="store_true", dest="report_success",
                            help="Send comment to review system on build success (in addition to vote up)")
        parser.add_argument("--report-only-fails", "-rof", action="store_true", dest="only_fails",
                            help="Include only the list of failed steps to reporting comments")
        parser.add_argument("--report-no-vote", "-rnv", action="store_true", dest="no_vote",
                            help="Do not vote up/down review depending on result")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.observers = []
        self.report_initialized = False
        self.blocks_to_report = []
        self.artifacts_to_report = []
        self.code_report_comments = defaultdict(list)

        self.automation_server = self.automation_server_factory()

    def subscribe(self, observer):
        self.observers.append(observer)

    def report_review_link(self):
        for observer in self.observers:
            self.out.log("Review can be found here: " + observer.get_review_link())

    @make_block("Reporting build start", pass_errors=False)
    def report_build_started(self):
        self.report_initialized = True
        if not self.observers:
            self.out.log("Nowhere to report. Skipping...")
            return

        if not self.settings.report_start:
            self.out.log("Reporting skipped. To report build start, use '--report-build-start' option")
            return

        text = "Review update detected!\n\n"
        text += self.automation_server.report_build_location()
        text += "\n\nPlease do not submit until revision testing is finished."

        for observer in self.observers:
            observer.report_start(text)

    def add_block_to_report(self, block):
        self.blocks_to_report.append(block)

    def report_artifacts(self, artifact_list):
        self.artifacts_to_report.extend(artifact_list)

    def code_report(self, path, message):
        self.code_report_comments[path].append(message)

    @make_block("Reporting build result", pass_errors=False)
    def report_build_result(self):
        if self.report_initialized is False:
            self.out.log("Not reporting: no build steps executed")
            return

        is_successful = True
        text = "Here is the summarized build result:\n"
        self.out.log(text)
        for step in self.blocks_to_report:
            text, status = self._report_steps_recursively(step, text, "")
            is_successful = is_successful and status
        if self.settings.only_fails and is_successful:
            text += "  All steps succeeded"
            self.out.log("  All steps succeeded")

        if not self.observers:
            self.out.log("Nowhere to report. Skipping...")
            return

        if is_successful:
            self.out.log("Reporting successful build...")
            if not self.settings.report_success:
                text = "Sending comment skipped. " + \
                       "To report build success, use '--report-build-success' option"
                self.out.log(text)
                text = None
        else:
            self.out.log("Reporting failed build...")

        if text is not None:
            if not self.settings.report_start:
                text += "\n\n" + self.automation_server.report_build_location()

            if self.artifacts_to_report:
                text += "\n\nThe following artifacts were generated during check:\n"
                for item in self.artifacts_to_report:
                    text += "* " + item + "\n"
                text += "Please take a look."

            for observer in self.observers:
                observer.report_result(is_successful, text, no_vote=self.settings.no_vote)

        if self.code_report_comments:
            self.out.log("Reporting code report issues ")
            for observer in self.observers:
                observer.code_report_to_review(self.code_report_comments)

    def _report_steps_recursively(self, block, text, indent):
        if not self.settings.only_fails:
            text += indent + str(block) + '\n'
            self.out.report_step(indent + str(block), block.status)
        elif not block.is_successful():
            text += str(block) + '\n'
            self.out.report_step(str(block), block.status)

        is_successful = block.is_successful()
        for substep in block.children:
            text, status = self._report_steps_recursively(substep, text, indent + "  ")
            is_successful = is_successful and status
        return text, is_successful



import locale
import sys

from .base_output import BaseOutput

__all__ = [
    "TerminalBasedOutput"
]


class Colors:
    red = "\033[1;31m"
    dark_red = "\033[0;31m"
    green = "\033[1;32m"
    blue = "\033[1;34m"
    dark_cyan = "\033[0;36m"
    dark_yellow = "\033[0;33m"
    reset = "\033[00m"


def stdout(*args, **kwargs):
    sys.stdout.write(''.join(args))
    if not kwargs.get("no_enter", False):
        sys.stdout.write('\n')


class TerminalBasedOutput(BaseOutput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.block_level = 0
        self.unicode_acceptable = (locale.getpreferredencoding() == "UTF-8")

    def indent(self):
        for x in range(0, self.block_level):
            stdout("  " * x, " |   ", no_enter=True)

    def print_lines(self, *args):
        result = ''.join(args)
        lines = result.splitlines(False)
        for line in lines:
            self.indent()
            stdout(line)

    def open_block(self, num_str, name):
        self.indent()
        stdout(num_str, ' ', Colors.blue, name, Colors.reset)
        self.block_level += 1

    def close_block(self, num_str, name, status):
        self.block_level -= 1
        self.indent()
        if self.unicode_acceptable:
            block_end = " \u2514 "
        else:
            block_end = " | "

        if status == "Failed":
            stdout(self.block_level * "  ", block_end, Colors.red, "[Failed]", Colors.reset)
        else:
            stdout(self.block_level * "  ", block_end, Colors.green, "[Success]", Colors.reset)
        self.indent()
        stdout()

    def report_error(self, description):
        pass

    def report_skipped(self, message):
        self.print_lines(Colors.dark_cyan, message, Colors.reset)

    def report_step(self, message, status):
        color = Colors.red
        if status.lower() == "success":
            color = Colors.green

        if message.endswith(status):
            message = message[:-len(status)]
            message += color + status + Colors.reset
        self.log(message)

    def change_status(self, message):
        pass

    def log_exception(self, line):
        self.print_lines(Colors.dark_red, "Error: ", Colors.reset, line)

    def log_stderr(self, line):
        self.print_lines(Colors.dark_yellow, "stderr: ", Colors.reset, line)

    def log(self, line):
        self.print_lines("==> ", line)

    def log_external_command(self, command):
        self.print_lines("$ ", command)

    def log_shell_output(self, line):
        self.print_lines(line)



from ...lib.gravity import Module


class BaseOutput(Module):
    """
    Abstract base class for output drivers
    """

    def open_block(self, num_str, name):
        raise NotImplementedError

    def close_block(self, num_str, name, status):
        raise NotImplementedError

    def report_error(self, description):
        raise NotImplementedError

    def report_skipped(self, message):
        raise NotImplementedError

    def report_step(self, message, status):
        self.log(message)

    def change_status(self, message):
        raise NotImplementedError

    def log_exception(self, line):
        raise NotImplementedError

    def log_stderr(self, line):
        raise NotImplementedError

    def log(self, line):
        raise NotImplementedError

    def log_external_command(self, command):
        raise NotImplementedError

    def log_shell_output(self, line):
        raise NotImplementedError



import glob
import json
import os
from copy import deepcopy
from typing import cast, Dict, List, Optional, TextIO, Tuple

from ..configuration_support import Configuration, Step
from .output import HasOutput
from .project_directory import ProjectDirectory
from . import artifact_collector, reporter
from ..lib import utils
from ..lib.gravity import Dependency
from ..lib.utils import make_block
from .structure_handler import HasStructure


class CodeReportCollector(ProjectDirectory, HasOutput, HasStructure):
    reporter_factory = Dependency(reporter.Reporter)
    artifacts_factory = Dependency(artifact_collector.ArtifactCollector)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.artifacts: artifact_collector.ArtifactCollector = self.artifacts_factory()
        self.reporter: reporter.Reporter = self.reporter_factory()
        self.report_path: str = ""
        self.repo_diff: Optional[List[Tuple[Optional[str], Optional[str], Optional[str]]]]

    def prepare_environment(self, project_config: Configuration) -> Configuration:
        afterall_steps: Configuration = Configuration()
        for item in project_config.configs:
            if not item.code_report:
                continue
            if not self.report_path:
                self.report_path = os.path.join(self.settings.project_root, "code_report_results")
                if not os.path.exists(self.report_path):
                    os.makedirs(self.report_path)
            temp_filename: str = "${CODE_REPORT_FILE}"
            name: str = utils.calculate_file_absolute_path(self.report_path, item.name) + ".json"
            actual_filename: str = os.path.join(self.report_path, name)
            item.replace_string(temp_filename, actual_filename)
            afterall_steps += [deepcopy(item)]
        return afterall_steps

    @make_block("Processing code report results")
    def report_code_report_results(self) -> None:
        reports: List[str] = glob.glob(self.report_path + "/*.json")
        for report_file in reports:
            with open(report_file, "r") as f:
                text: str = f.read()
                report: Optional[List[Dict[str, str]]] = None
                if text:
                    report = json.loads(text)

            json_file: TextIO = self.artifacts.create_text_file("Static_analysis_report.json")
            json_file.write(json.dumps(report, indent=4))

            if report:
                for result in report:
                    text = result["symbol"] + ": " + result["message"]
                    self.reporter.code_report(result["path"], {"message": text, "line": result["line"]})

            if report:
                text = str(len(report)) + " issues"
                self.out.log_stderr("Found " + text)
                self.out.report_build_status(os.path.splitext(os.path.basename(report_file))[0] + ": " + text)
            elif report == []:
                self.out.log("Issues not found.")
            else:  # if nothing was written to file
                self.out.log_stderr("There are no results in code report file. Something went wrong.")



import os
import shutil

from ..error_state import HasErrorState
from ...lib.ci_exception import CriticalCiException
from ...lib.utils import make_block
from ...lib import utils
from ..output import HasOutput
from ..structure_handler import HasStructure
from . import base_vcs

__all__ = [
    "LocalMainVcs"
]


class LocalMainVcs(base_vcs.BaseDownloadVcs, HasOutput, HasStructure, HasErrorState):
    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Local files",
                                                     "Parameters for file settings in case of no VCS used")

        parser.add_argument('--file-source-dir', '-fsd', dest="source_dir", metavar='SOURCE_DIR',
                            help="A local folder for project sources to be copied from. "
                                 "This option is only needed when '--driver-type' is set to 'none'")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not getattr(self.settings, "source_dir", None):
            self.error("""The source directory is not specified.
            
                          The source directory defines the location of the project sources on a local
                          filesystem.
            
                          Please specify source directory by using --file-source-dir
                          command-line option or SOURCE_DIR environment variable.""")
        else:
            self.source_dir = utils.parse_path(self.settings.source_dir, os.getcwd())

    def calculate_file_diff(self):  # pylint: disable=no-self-use
        return {}                   # No file diff can be calculated for local VCS

    @make_block("Copying sources to working directory")
    def prepare_repository(self):
        self.sources_need_cleaning = True        # pylint: disable=attribute-defined-outside-init
        try:
            self.out.log("Moving sources to '" + self.settings.project_root + "'...")
            shutil.copytree(self.source_dir, self.settings.project_root)
            self.append_repo_status("Got sources from: " + self.source_dir + "\n")
        except OSError as e:
            text = f"{e}\nPossible reasons of this error:"
            text += f"\n * Sources path, passed to the script ('{self.settings.source_dir}')," + \
                    " does not lead to actual sources or was processed incorrectly"
            text += "\n * Directory '{}' already exists in working dir (e.g. due to previous builds)".format(
                os.path.basename(self.settings.project_root))
            text += "\n * File reading permissions troubles"
            raise CriticalCiException(text) from e



import sys
from types import TracebackType
from typing import ClassVar, Optional

from ...lib.gravity import Module, Dependency
from ...lib import utils
from .base_output import BaseOutput
from .terminal_based_output import TerminalBasedOutput
from .teamcity_output import TeamcityOutput

__all__ = [
    "HasOutput",
    "MinimalOut",
    "Output"
]


class Output(Module):
    teamcity_driver_factory = Dependency(TeamcityOutput)
    terminal_driver_factory = Dependency(TerminalBasedOutput)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Output", "Log appearance parameters")
        parser.add_argument("--out-type", "-ot", dest="type", choices=["tc", "term", "jenkins"],
                            help="Type of output to produce (tc - TeamCity, jenkins - Jenkins, term - terminal). "
                                 "TeamCity and Jenkins environments are detected automatically when launched on build "
                                 "agent.")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.driver: BaseOutput = \
            utils.create_driver(local_factory=self.terminal_driver_factory,
                                teamcity_factory=self.teamcity_driver_factory,
                                jenkins_factory=self.terminal_driver_factory,
                                env_type=self.settings.type)

    def log(self, line: str) -> None:
        self.driver.log(line)

    def log_external_command(self, command: str) -> None:
        self.driver.log_external_command(command)

    def open_block(self, number: str, name: str) -> None:
        self.driver.open_block(number, name)

    def close_block(self, number: str, name: str, status: str) -> None:
        self.driver.close_block(number, name, status)

    def report_build_status(self, status: str) -> None:
        self.driver.change_status(status)

    # TODO: pass build problem to the Report module
    def report_build_problem(self, problem: str) -> None:
        self.driver.report_error(problem)

    def report_skipped(self, message: str) -> None:
        self.driver.report_skipped(message)

    def report_step(self, message: str, status: str) -> None:
        self.driver.report_step(message, status)

    def log_exception(self, line: str) -> None:
        self.driver.log_exception(line)

    def log_stderr(self, line: str) -> None:
        self.driver.log_stderr(line)

    def log_shell_output(self, line: str) -> None:
        self.driver.log_shell_output(line)


class HasOutput(Module):
    out_factory: ClassVar = Dependency(Output)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.out: Output = self.out_factory()


class MinimalOut:
    def __init__(self, silent=False):
        self.silent = silent

    def log(self, line: str) -> None:
        if not self.silent:
            print(line)

    @staticmethod
    def report_build_problem(problem: str) -> None:
        pass

    @staticmethod
    def log_exception(exc: Exception) -> None:
        ex_traceback: Optional[TracebackType] = sys.exc_info()[2]
        sys.stderr.write("Unexpected error.\n" + utils.format_traceback(exc, ex_traceback))



from typing import Union
import os

from ..error_state import HasErrorState
from ...lib.module_arguments import ModuleArgumentParser
from ...lib import utils
from ..output import HasOutput
from .base_server import BaseServerForHostingBuild, BaseServerForTrigger

__all__ = [
    "JenkinsServerForTrigger",
    "JenkinsServerForHostingBuild"
]


class JenkinsServerForTrigger(HasOutput, BaseServerForTrigger, HasErrorState):

    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        parser = argument_parser.get_or_create_group("Jenkins variables", "Jenkins-specific parameters")
        parser.add_argument('--jenkins-trigger-url', '-jtu', dest='trigger_url',
                            help='Url to trigger, must include exactly one conversion specifier (%%s) to be '
                                 'replaced by CL number, for example: http://localhost/%%s', metavar="URL")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.check_required_option("trigger_url", """
            The Jenkins url for triggering build is not specified.
            
            Please specify the url by using '--jenkins-trigger-url' ('-jtu') command-line
            option or URL environment variable.
            """)

    def trigger_build(self, revision: str) -> None:
        processed_url = self.settings.trigger_url % revision
        self.out.log("Triggering url %s" % processed_url)
        utils.make_request(processed_url)


class JenkinsServerForHostingBuild(BaseServerForHostingBuild, HasErrorState):

    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        parser = argument_parser.get_or_create_group("Jenkins variables", "Jenkins-specific parameters")
        parser.add_argument("--jenkins-build-url", "-jbu", dest="build_url", metavar="BUILD_URL",
                            help="Link to build on Jenkins (automatically set by Jenkins)")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.check_required_option("build_url", """
            The Jenkins url of the ongoing build is not specified
            
            Please specify the url by using '--jenkins-build-url' ('-jbu') command-line
            option or BUILD_URL environment variable.
            """)

    def report_build_location(self) -> str:
        log_link: str = self.settings.build_url + "console"
        return "Here is the link to build log on Jenkins: " + log_link

    def artifact_path(self, local_artifacts_dir: str, item: str) -> str:
        artifact_link: str = self.settings.build_url + "artifact/" + \
                             os.path.relpath(local_artifacts_dir, os.getcwd()) + "/"
        return artifact_link + item



import argparse
import os
import sys

import sh
from lxml import etree

from . import utils


class SvaceAnalyzer:

    @staticmethod
    def define_arguments():
        parser = argparse.ArgumentParser(description="Svace analyzer")
        parser.add_argument("--build-cmd", dest="build_cmd", nargs='+',
                            help="Relative path to build script or command itself")
        parser.add_argument("--lang", dest="lang", choices=["JAVA", "CXX"], help="Language to analyze")
        parser.add_argument("--project-name", dest="project_name", help="Svace project name defined on server")
        utils.add_common_arguments(parser)
        return parser

    def __init__(self, settings):
        self.settings = settings

        self.settings.svace_home = "/opt/svace-x64-linux"
        self.project_name = self.settings.project_name + "_" + self.settings.lang

        self.work_folder = os.path.join(os.getcwd(), self.project_name)
        self.settings.jack_option = "--enable-jack-interception"
        self.settings.hash_server_memory = "2048"
        self.settings.verbose = "--verbose"

        if self.settings.lang.upper() == "JAVA":
            self.disabled_language = "cxx"
            self.enabled_language = "java"
        elif self.settings.lang.upper() == "CXX":
            self.disabled_language = "java"
            self.enabled_language = "cxx"
        self.json_file = settings.result_file

    def analyze(self):
        try:
            cmd_analyze = sh.Command(self.settings.svace_home + "/bin/svace")
            cmd_analyze("analyze", "-q", "--svace-dir", self.work_folder, "--disable-language",
                        self.disabled_language, "--enable-language", self.enabled_language,
                        "--preset", self.enabled_language)

            svres_full = os.path.join(self.work_folder, "analyze-res", self.project_name + ".svres")
            root = etree.parse(svres_full)

            warn_info = root.xpath('//WarnInfo')
            issues_loads = []
            for info in warn_info:
                issue = dict()
                issue["symbol"] = info.attrib["warnClass"]
                issue["message"] = "\nWarning message: " + info.attrib["msg"]
                issue["path"] = info.attrib["file"]
                issue["line"] = info.attrib["line"]
                issues_loads.append(issue)
            utils.analyzers_output(self.json_file, issues_loads)
            if issues_loads:
                return 1
        except etree.XMLSyntaxError as e:
            sys.stderr.write(e.error_log) #TODO: check is it corerct
            return 2
        except Exception as e:
            sys.stderr.write(str(e))
            return 2
        return 0

    def execute(self):
        if not os.path.exists(self.settings.svace_home):
            sys.stderr.write("SVACE_HOME=" + self.settings.svace_home + " folder doesn't exist.")
            return 2

        try:
            cmd_init = sh.Command(self.settings.svace_home + "/bin/svace")
            cmd_init("init", "-f", "--shared=all", "--bare", self.work_folder)
            cmd_build = sh.Command(self.settings.svace_home + "/bin/svace")
            cmd_build("build", self.settings.jack_option, self.settings.verbose,
                      "--hash-server-memory", self.settings.hash_server_memory,
                      "--svace-dir", self.work_folder, self.settings.build_cmd,
                      _err=sys.stderr, _out=sys.stdout)
        except sh.ErrorReturnCode_255:
            sys.stderr.write("Svace exited with error code 255. No build object found.\n")
            return 2
        except Exception as e:
            sys.stderr.write(str(e))
            return 2
        return self.analyze()


def form_arguments_for_documentation():
    return SvaceAnalyzer.define_arguments()


def main():
    analyzer_namespace = SvaceAnalyzer.define_arguments().parse_args()
    analyze = SvaceAnalyzer(analyzer_namespace)
    return analyze.execute()


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)



import json
import sys


def add_common_arguments(parser):
    parser.add_argument("--result-file", dest="result_file",
                        help="File for storing json results of Universum run. Set it to \"${CODE_REPORT_FILE}\" "
                             "for running from Universum, variable will be handled during run. If you run this "
                             "script separately from Universum, just name the result file or leave it empty.")


def analyzers_output(json_file: str, issues_loads) -> None:
    issues = json.dumps(issues_loads, indent=4)
    if json_file:
        open(json_file, "w").write(issues)
    else:
        sys.stdout.write(issues)



import argparse
import glob
import json
import sys
import subprocess

from . import utils


class PylintAnalyzer:
    """
    Pylint runner.
    Specify parameters such as project folders, config file for code report tool.
    For example:
    universum_pylint --python-version 2 --files *.py tests/
    ./pylint.py --python-version 2 --files *.py tests/
    Output: json of the found issues in the code.
    """

    @staticmethod
    def define_arguments():
        parser = argparse.ArgumentParser(description="Pylint analyzer")
        parser.add_argument("--files", dest="file_list", nargs='+', required=True,
                            help="Python files and Python packages for Pylint. "
                                 "Files could be defined as a single python file"
                                 " *.py or directories with __init__.py file in the directory.")
        parser.add_argument("--rcfile", dest="rcfile", type=str, help="Specify a configuration file.")
        parser.add_argument("--python-version", dest="version", default="3",
                            help="Version of the python interpreter, such as 2, 3 or 3.7. "
                                 "Pylint analyzer uses this parameter to select python binary for launching pylint. "
                                 "For example, if the version is 3.7, it uses the following command: "
                                 "'python3.7 -m pylint <...>'")
        utils.add_common_arguments(parser)
        return parser

    def __init__(self, settings):
        self.settings = settings

    def execute(self):
        cmd = [f"python{self.settings.version}", '-m', 'pylint', '-f', 'json']
        if self.settings.rcfile:
            cmd.append(f'--rcfile={self.settings.rcfile}')

        for pattern in self.settings.file_list:
            cmd.extend(glob.glob(pattern))

        result = subprocess.run(cmd, universal_newlines=True,  # pylint: disable=subprocess-run-check
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if result.stderr and not result.stdout:
            sys.stderr.write(result.stderr)
            return result.returncode

        return self._handle_pylint_output(result.stdout)

    def _handle_pylint_output(self, issues: str) -> int:
        try:
            loads = json.loads(issues)
            issues_loads = []
            for issue in loads:
                # pylint has its own escape rules for json output of "message" values.
                # it uses cgi.escape lib and escapes symbols <>&
                issue["message"] = issue["message"].replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")
                issues_loads.append(issue)

            utils.analyzers_output(self.settings.result_file, issues_loads)
            if issues_loads:
                return 1
            return 0

        except ValueError as e:
            sys.stderr.write(str(e))
            sys.stderr.write("The following string produced by the pylint launch cannot be parsed as JSON:\n")
            sys.stderr.write(issues)
            return 2


def form_arguments_for_documentation():
    return PylintAnalyzer.define_arguments()


def main():
    analyzer_namespace = PylintAnalyzer.define_arguments().parse_args()
    analyze = PylintAnalyzer(analyzer_namespace)
    return analyze.execute()


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)



from typing import ClassVar
from . import __title__
from .lib.ci_exception import SilentAbortException
from .lib.gravity import Dependency
from .lib.module_arguments import ModuleArgumentParser
from .modules import vcs, artifact_collector, reporter, launcher, code_report_collector
from .modules.output import HasOutput
from .configuration_support import Configuration


__all__ = ["Main"]


class Main(HasOutput):
    description: ClassVar[str] = __title__
    vcs_factory: ClassVar = Dependency(vcs.MainVcs)
    launcher_factory: ClassVar = Dependency(launcher.Launcher)
    artifacts_factory: ClassVar = Dependency(artifact_collector.ArtifactCollector)
    reporter_factory: ClassVar = Dependency(reporter.Reporter)
    code_report_collector_factory: ClassVar = Dependency(code_report_collector.CodeReportCollector)

    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        argument_parser.add_hidden_argument("--no-finalize", action="store_true", dest="no_finalize",
                                            help="Skip 'Finalizing' step: "
                                                 "do not clear sources, do not revert workspace vcs, etc. "
                                                 "Is applied automatically when using existing VCS client")

        argument_parser.add_hidden_argument("--finalize-only", action="store_true", dest="finalize_only",
                                            help="Perform only 'Finalizing' step: "
                                                 "clear sources, revert workspace vcs, etc. "
                                                 "Recommended to use after '--no-finalize' runs. "
                                                 "Please make sure to move artifacts from working directory "
                                                 "or pass different artifact folder")

        argument_parser.add_hidden_argument("--clean-build", action="store_true", dest="clean_build",
                                            help="Clean artifact and build directory before build "
                                                 "instead of raising exception. Not recommended for CI configurations!")

        argument_parser.add_argument("--build-only-latest", action="store_true", dest="build_only_latest",
                                     help="Skip build if review version isn't latest")
        argument_parser.add_argument("--no-diff", action="store_true", dest="no_diff",
                                     help="Only applies to build steps where ``code_report=True``; "
                                          "disables calculating analysis diff for changed files, "
                                          "in this case full analysis report will be published")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.vcs: vcs.MainVcs = self.vcs_factory()
        self.launcher: launcher.Launcher = self.launcher_factory()
        self.artifacts: artifact_collector.ArtifactCollector = self.artifacts_factory()
        self.reporter: reporter.Reporter = self.reporter_factory()
        self.code_report_collector: code_report_collector.CodeReportCollector = self.code_report_collector_factory()

    def execute(self) -> None:
        if self.settings.clean_build:
            self.vcs.clean_sources_silently()
            self.artifacts.clean_artifacts_silently()

        if self.settings.finalize_only:
            self.vcs.driver.sources_need_cleaning = True
            self.out.log("Execution skipped because of '--finalize-only' option")
            return

        self.reporter.report_review_link()
        if self.settings.build_only_latest:
            if not self.vcs.is_latest_review_version():
                self.out.log("Build skipped because review revision is not latest")
                self.out.report_build_status("Skipped - review revision is not latest")
                raise SilentAbortException(application_exit_code=0)

        self.vcs.prepare_repository()
        project_configs: Configuration = self.launcher.process_project_configs()
        afterall_configs: Configuration = self.code_report_collector.prepare_environment(project_configs)
        self.artifacts.set_and_clean_artifacts(project_configs)

        self.reporter.report_build_started()
        self.launcher.launch_project()
        if afterall_configs:
            if not self.settings.no_diff:
                try:
                    repo_diff = self.vcs.revert_repository()
                except NotImplementedError:
                    self.out.log("Diff calculation for code report is skipped because current VCS doesn't support it")
                else:
                    self.launcher.launch_custom_configs(afterall_configs)
                    self.code_report_collector.repo_diff = repo_diff
            self.code_report_collector.report_code_report_results()
        self.artifacts.collect_artifacts()
        self.reporter.report_build_result()

    def finalize(self) -> None:
        if self.settings.no_finalize:
            self.out.log("Cleaning skipped because of '--no-finalize' option")
            return
        self.vcs.finalize()



from typing import Any, Callable, List, Optional, TypeVar, Union
from types import TracebackType
import os
import sys
import traceback

import requests
from requests.models import Response

from .ci_exception import CiException, CriticalCiException, SilentAbortException
from .gravity import Module

__all__ = [
    "strip_path_start",
    "parse_path",
    "calculate_file_absolute_path",
    "detect_environment",
    "create_driver",
    "format_traceback",
    "catch_exception",
    "trim_and_convert_to_unicode",
    "convert_to_str",
    "unify_argument_list",
    "Uninterruptible",
    "make_block",
    "make_request"
]

ReturnT = TypeVar('ReturnT')
DecoratorT = Callable[[Callable[..., ReturnT]], Callable[..., ReturnT]]


def strip_path_start(line: str) -> str:
    if line.startswith("./"):
        return line[2:]
    return line


def parse_path(path: str, starting_point: str) -> str:
    if path.startswith('/'):
        path = os.path.join(path)
    else:
        path = os.path.join(starting_point, path)

    return os.path.abspath(path)


def calculate_file_absolute_path(target_directory: str, file_basename: str) -> str:
    name = file_basename.replace(" ", "_")
    name = name.replace("/", "\\")
    if name.startswith('_'):
        name = name[1:]
    return os.path.join(target_directory, name)


def detect_environment() -> str:
    """
    :return: "tc" if the script is launched on TeamCity agent,
             "jenkins" is launched on Jenkins agent,
             "terminal" otherwise
    """
    teamcity = "TEAMCITY_VERSION" in os.environ
    jenkins = "JENKINS_HOME" in os.environ
    pycharm = "PYCHARM_HOSTED" in os.environ
    if pycharm:
        return "terminal"
    if teamcity and not jenkins:
        return "tc"
    if not teamcity and jenkins:
        return "jenkins"
    return "terminal"


LocalFactoryT = TypeVar('LocalFactoryT', bound=Module)
TeamcityFactoryT = TypeVar('TeamcityFactoryT', bound=Module)
JenkinsFactoryT = TypeVar('JenkinsFactoryT', bound=Module)


def create_driver(local_factory: Callable[[], LocalFactoryT],
                  teamcity_factory: Callable[[], TeamcityFactoryT],
                  jenkins_factory: Callable[[], JenkinsFactoryT],
                  env_type: str = "") -> Union[LocalFactoryT, TeamcityFactoryT, JenkinsFactoryT]:
    if not env_type:
        env_type = detect_environment()

    if env_type == "tc":
        return teamcity_factory()
    if env_type == "jenkins":
        return jenkins_factory()
    return local_factory()


def format_traceback(exc: Exception, trace: Optional[TracebackType]) -> str:
    tb_lines: List[str] = traceback.format_exception(exc.__class__, exc, trace)
    tb_text: str = ''.join(tb_lines)
    return tb_text


def catch_exception(exception_name: str, ignore_if: str = None) -> DecoratorT:
    def decorated_function(function):
        def function_to_run(*args, **kwargs):
            result: ReturnT = None
            try:
                result = function(*args, **kwargs)
                return result
            except Exception as e:
                if not type(e).__name__ == exception_name:
                    raise
                if ignore_if is not None:
                    if ignore_if in str(e):
                        return result
                raise CriticalCiException(str(e)) from e
        return function_to_run
    return decorated_function


def trim_and_convert_to_unicode(line: Union[bytes, str]) -> str:
    if isinstance(line, bytes):
        line = line.decode("utf-8")
    elif not isinstance(line, str):
        line = str(line)

    if line.endswith("\n"):
        line = line[:-1]

    return line


def convert_to_str(line: Union[bytes, str]) -> str:
    if isinstance(line, bytes):
        return line.decode("utf8", "replace")
    return str(line)


def unify_argument_list(source_list: Optional[List[str]], separator: str = ',',
                        additional_list: Optional[List[str]] = None) -> List[str]:
    resulting_list: List[str] = additional_list if additional_list else []

    # Add arguments parsed by ModuleArgumentParser, including list elements generated by nargs='+'
    if source_list is not None:
        for item in source_list:
            if isinstance(item, list):
                resulting_list.extend(item)
            else:
                resulting_list.append(item)

    # Remove None and empty elements added by previous steps
    resulting_list = [item for item in resulting_list if item]

    # Split one-element arguments and merge to one list
    resulting_list = [item.strip() for entry in resulting_list for item in entry.strip('"\'').split(separator)]

    return resulting_list


class Uninterruptible:
    def __init__(self, error_logger: Callable[[str], None]) -> None:
        self.return_code: int = 0
        self.error_logger: Callable[[str], None] = error_logger
        self.exceptions: List[str] = []

    def __enter__(self) -> Callable[..., None]:
        def excepted_function(func: Callable[..., None], *args, **kwargs):
            try:
                func(*args, **kwargs)
            except SilentAbortException as e:
                self.return_code = max(self.return_code, e.application_exit_code)
            except (KeyboardInterrupt, SystemExit):
                self.error_logger("Interrupted from outer scope\n")
                self.return_code = 3
            except Exception as e:
                ex_traceback = sys.exc_info()[2]
                self.exceptions.append(format_traceback(e, ex_traceback))
                self.return_code = max(self.return_code, 2)
        return excepted_function

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.return_code == 2:
            for entry in self.exceptions:
                sys.stderr.write(entry)
        if self.return_code != 0:
            raise SilentAbortException(application_exit_code=self.return_code)


def make_block(block_name: str, pass_errors: bool = True) -> DecoratorT:
    def decorated_function(func):
        def function_in_block(self, *args, **kwargs):
            return self.structure.run_in_block(func, block_name, pass_errors, self, *args, **kwargs)
        return function_in_block
    return decorated_function


def make_request(url: str, request_method: str = "GET", critical: bool = True, **kwargs) -> Response:
    try:
        response: Response = requests.request(method=request_method, url=url, **kwargs)
        response.raise_for_status()
        return response
    except requests.RequestException as error:
        text = f"Error opening URL, got '{type(error).__name__}' with following message:\n{error}"
        if critical:
            raise CriticalCiException(text) from error
        raise CiException(text) from error



from .base_output import BaseOutput

__all__ = [
    "TeamcityOutput"
]


def escape(message):
    return message.replace('\r', '').replace('|', '||').replace('\'', '|\'').replace('[', '|[').replace(']', '|]')


class TeamcityOutput(BaseOutput):
    def open_block(self, num_str, name):
        print(u"##teamcity[blockOpened name='{} {}']".format(num_str, escape(name)))

    def close_block(self, num_str, name, status):
        print(u"##teamcity[blockClosed name='{} {}']".format(num_str, escape(name)))

    def report_error(self, description):
        print(u"##teamcity[buildProblem description='<{}>']".format(escape(description)))

    def report_skipped(self, message):
        lines = message.split("\n")
        for single_line in lines:
            print(u"##teamcity[message text='{}' status='WARNING']".format(escape(single_line)))

    def change_status(self, message):
        print(u"##teamcity[buildStatus text='{}']".format(escape(message)))

    def log_exception(self, line):
        lines = line.split("\n")
        for single_line in lines:
            print(u"##teamcity[message text='{}' status='ERROR']".format(escape(single_line)))

    def log_stderr(self, line):
        lines = line.split("\n")
        for single_line in lines:
            print(u"##teamcity[message text='{}' status='WARNING']".format(escape(single_line)))

    def log(self, line):
        print(u"==>", line)

    def log_external_command(self, command):
        print(u"$", command)

    def log_shell_output(self, line):
        print(line)



import argparse
import difflib
import sys
import os

import re
import sh

from . import utils

# The maximum number of lines to write separate comments for
# If exceeded, summarized comment will be provided instead
MAX_LINES = 11


def replace_invisible_symbols(line):
    for old_str, new_str in zip([u" ", u"\t", u"\n"], [u"\u00b7", u"\u2192\u2192\u2192\u2192", u"\u2193\u000a"]):
        return line.replace(old_str, new_str)


def add_files_recursively(item_path):
    files = []
    item_path = os.path.join(os.getcwd(), item_path)
    if os.path.isfile(item_path):
        files.append(item_path)
    elif os.path.isdir(item_path):
        for root_dir, _, file_names in os.walk(item_path):
            for file_name in file_names:
                files.append(os.path.join(root_dir, file_name))
    else:
        sys.stderr.write(item_path + " doesn't exist.")
        sys.exit(2)

    return files


def get_issue_json_format(src_file, diff):
    issue = dict()
    issue["symbol"] = "Uncrustify Code Style issue"
    # Check block size
    diff_size = len(diff["before"].splitlines())
    if diff_size > MAX_LINES:
        issue["message"] = "\nLarge block of code (" + str(diff_size) + \
                           " lines) has issues\n"
    else:
        # Message with before&after
        issue["message"] = "\nOriginal code:\n```diff\n" + diff["before"] + "```\n" + \
                           "Uncrustify generated code:\n```diff\n" + diff['after'] + "```\n"

    issue["path"] = os.path.relpath(src_file, os.getcwd())
    issue["line"] = diff["line"]
    return issue


def get_text_for_block(start, end, lines):
    return replace_invisible_symbols(''.join(lines[start: end]))


def get_mismatching_block(first_match, second_match, src_lines, fixed_lines):
    block = dict()
    first_match_end_in_source = first_match.a + first_match.size
    first_match_end_in_fixed = first_match.b + first_match.size
    second_match_start_in_source = second_match.a
    second_match_start_in_fixed = second_match.b
    if first_match_end_in_source != second_match_start_in_source:
        block['line'] = second_match_start_in_source
        block["before"] = get_text_for_block(first_match_end_in_source - 1, second_match_start_in_source, src_lines)
        block['after'] = get_text_for_block(first_match_end_in_fixed - 1, second_match_start_in_fixed, fixed_lines)
    return block


class UncrustifyAnalyzer:
    """
    Uncrustify runner.
    Specify parameters such as file list, config file for code report tool.
    For example: universum_uncrustify --files *.py tests/
    Output: json of the found issues in the code.
    """
    @staticmethod
    def define_arguments():
        parser = argparse.ArgumentParser(description="Uncrustify analyzer")
        parser.add_argument("--files", "-f", dest="file_names", nargs="*", default=[],
                            help="File or directory to check; accepts multiple values; "
                                 "all files specified by both '--files' and '--file-list' "
                                 "are gathered into one combined list of files")
        parser.add_argument("--file-list", "-fl", dest="file_lists", nargs="*", default=[],
                            help="Text file with list of files or directories to check; "
                                 "can be used with '--files'; accepts multiple values; "
                                 "all files specified by both '--files' and '--file-list' "
                                 "are gathered into one combined list of files")
        parser.add_argument("--cfg-file", "-cf", dest="cfg_file",
                            help="Name of the configuration file of Uncrustify; "
                                 "can also be set via 'UNCRUSTIFY_CONFIG' env. variable")
        parser.add_argument("--filter-regex", "-r", dest="pattern_form", nargs="*", default=[],
                            help="(optional) Python 2.7 regular expression filter to apply to "
                                 "combined list of files to check")
        parser.add_argument("--output-directory", "-od", dest="output_directory", default="uncrustify",
                            help="Directory to store fixed files, generated by Uncrustify "
                                 "and HTML files with diff; the default value is 'uncrustify'")

        utils.add_common_arguments(parser)
        return parser

    def __init__(self, settings):
        self.settings = settings
        self.wrapcolumn = None
        self.tabsize = None

    def parse_files(self):
        files = []
        file_lines = []
        for file_name in self.settings.file_names:
            files.extend(add_files_recursively(file_name))
        for file_list in self.settings.file_lists:
            with open(file_list) as f:
                for file_name in f.readlines():
                    file_lines.append(file_name.strip())
        for file_name in file_lines:
            files.extend(add_files_recursively(file_name))
        for pattern in self.settings.pattern_form:
            regexp = re.compile(pattern)
            files = [file_name for file_name in files if regexp.match(file_name)]
        files = [os.path.relpath(file_name) for file_name in files]
        if not files:
            sys.stderr.write("Please provide at least one file for analysis")
            return 2

        return files

    def get_htmldiff_parameters(self):
        if not (self.wrapcolumn and self.tabsize):
            with open(self.settings.cfg_file) as config:
                for line in config.readlines():
                    if line.startswith("code_width"):
                        self.wrapcolumn = int(line.split()[2])
                    if line.startswith("input_tab_size"):
                        self.tabsize = int(line.split()[2])
        return self.wrapcolumn, self.tabsize

    def generate_html_diff(self, file_name, left_lines, right_lines):
        wrapcolumn, tabsize = self.get_htmldiff_parameters()
        differ = difflib.HtmlDiff(tabsize=tabsize, wrapcolumn=wrapcolumn)
        file_name = os.path.relpath(file_name, self.settings.output_directory).replace('/', '_') + '.html'
        with open(os.path.join(self.settings.output_directory, file_name), 'w') as outfile:
            outfile.write(differ.make_file(left_lines, right_lines, context=False))

    def get_file_issues(self, src_file):
        self.settings.output_directory = os.path.join(os.getcwd(), self.settings.output_directory)
        # Uncrustify copies absolute path in its target folder, that's why we use '+'
        uncrustify_file = os.path.normpath(self.settings.output_directory + '/' + src_file)

        with open(src_file) as src:
            src_lines = src.readlines()
        with open(uncrustify_file) as fixed:
            fixed_lines = fixed.readlines()

        file_issues = []
        matching_blocks = difflib.SequenceMatcher(a=src_lines, b=fixed_lines).get_matching_blocks()
        previous_match = matching_blocks[0]
        for match in matching_blocks[1:]:
            block = get_mismatching_block(previous_match, match, src_lines, fixed_lines)
            previous_match = match
            if block:
                file_issues.append(get_issue_json_format(src_file, block))

        # Generate html diff
        if file_issues:
            self.generate_html_diff(uncrustify_file, src_lines, fixed_lines)

        return file_issues

    def execute(self):
        if not self.settings.cfg_file and ('UNCRUSTIFY_CONFIG' not in os.environ):
            sys.stderr.write("Please specify the '--cfg_file' parameter "
                             "or set an env. variable 'UNCRUSTIFY_CONFIG'")
            return 2

        files = self.parse_files()
        try:
            cmd = sh.Command("uncrustify")
            cmd("-c", self.settings.cfg_file, "--prefix", self.settings.output_directory, files)
        except sh.ErrorReturnCode as e:
            sys.stderr.write(str(e) + '\n')

        issues_loads = []
        for file_name in files:
            issues_loads.extend(self.get_file_issues(file_name))
        if issues_loads:
            utils.analyzers_output(self.settings.result_file, issues_loads)
            return 1

        return 0


def form_arguments_for_documentation():
    return UncrustifyAnalyzer.define_arguments()


def main():
    analyzer_namespace = UncrustifyAnalyzer.define_arguments().parse_args()
    analyze = UncrustifyAnalyzer(analyzer_namespace)
    return analyze.execute()


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)



from .lib import utils
from .lib.gravity import Dependency
from .lib.utils import make_block
from .modules import vcs
from .modules.error_state import HasErrorState
from .modules.output import HasOutput
from .modules.structure_handler import HasStructure

__all__ = ["Submit"]


class Submit(HasOutput, HasStructure, HasErrorState):
    description = "Submitting module of Universum"
    vcs_factory = Dependency(vcs.SubmitVcs)

    @staticmethod
    def define_arguments(parser):
        parser.add_argument("--create-review", action="store_true", dest="review",
                            help="create deletable review (shelve for P4, temp branch for Git) "
                                 "instead of actual submitting to repo")
        parser.add_argument("--edit-only", action="store_true", dest="edit_only",
                            help="Only submit existing vcs modifications, no adding or deleting")

        parser.add_argument('--commit-message', '-cm', dest='commit_message', metavar="COMMIT_MESSAGE",
                            help='Commit message to add')
        parser.add_argument("--reconcile-list", "-rl", action="append", nargs='+', dest="reconcile_list",
                            metavar="RECONCILE_LIST",
                            help="List of vcs or directories to be reconciled for commit. "
                                 "Relative paths starting at client root are supported")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.check_required_option("commit_message", """
            Commit message is not specified.
            
            Please use '--commit-message' option or COMMIT_MESSAGE environment variable.
            """)

        self.vcs = self.vcs_factory()
        self.client = None

    @make_block("Executing")
    def execute(self):
        path_list = utils.unify_argument_list(self.settings.reconcile_list)
        change = self.vcs.driver.submit_new_change(self.settings.commit_message,
                                                   path_list,
                                                   review=self.settings.review,
                                                   edit_only=self.settings.edit_only)

        if change == 0:
            self.out.log("Nothing to submit")
        elif self.settings.review:
            self.out.log("Review commit " + change + " created")
        else:
            self.out.log("Change " + change + " submitted")

    @make_block("Finalizing", pass_errors=False)
    def finalize(self):
        self.vcs.finalize()
